#include "main.h"
#include <string.h>

#define CONFIG_GROUP_APP "application"
#define CONFIG_GROUP_APP_ENABLE_PERF_MEASUREMENT "enable-perf-measurement"
#define CONFIG_GROUP_APP_PERF_MEASUREMENT_INTERVAL "perf-measurement-interval-sec"

GST_DEBUG_CATEGORY_EXTERN(APP_CFG_PARSER_CAT);

#define CHECK_ERROR(error)                                   \
  if (error)                                                 \
  {                                                          \
    GST_CAT_ERROR(APP_CFG_PARSER_CAT, "%s", error->message); \
    goto done;                                               \
  }

NvDsSourceConfig global_source_config;

static gboolean
parse_source_list(NvDsConfig *config, GKeyFile *key_file,
                  gchar *cfg_file_path)
{
  gboolean ret = FALSE;
  gchar **keys = NULL;
  gchar **key = NULL;
  GError *error = NULL;
  gsize num_strings;

  keys = g_key_file_get_keys(key_file, CONFIG_GROUP_SOURCE_LIST, NULL, &error);
  CHECK_ERROR(error);

  for (key = keys; *key; key++)
  {
    if (!g_strcmp0(*key, CONFIG_GROUP_SOURCE_LIST_NUM_SOURCE_BINS))
    {
      config->total_num_sources =
          g_key_file_get_integer(key_file, CONFIG_GROUP_SOURCE_LIST,
                                 CONFIG_GROUP_SOURCE_LIST_NUM_SOURCE_BINS, &error);
      CHECK_ERROR(error);
    }
    else if (!g_strcmp0(*key, CONFIG_GROUP_SOURCE_LIST_URI_LIST))
    {
      config->uri_list =
          g_key_file_get_string_list(key_file, CONFIG_GROUP_SOURCE_LIST,
                                     CONFIG_GROUP_SOURCE_LIST_URI_LIST, &num_strings, &error);
      if (num_strings > MAX_SOURCE_BINS)
      {
        NVGSTDS_ERR_MSG_V("App supports max %d sources", MAX_SOURCE_BINS);
        goto done;
      }
      CHECK_ERROR(error);
    }
    else
    {
      NVGSTDS_WARN_MSG_V("Unknown key '%s' for group [%s]", *key,
                         CONFIG_GROUP_SOURCE_LIST);
    }
  }

  if (g_key_file_has_key(key_file, CONFIG_GROUP_SOURCE_LIST,
                         CONFIG_GROUP_SOURCE_LIST_URI_LIST, &error))
  {
    if (g_key_file_has_key(key_file, CONFIG_GROUP_SOURCE_LIST,
                           CONFIG_GROUP_SOURCE_LIST_NUM_SOURCE_BINS, &error))
    {
      if (num_strings != config->total_num_sources)
      {
        NVGSTDS_ERR_MSG_V("Mismatch in URIs provided and num-source-bins.");
        goto done;
      }
    }
    else
    {
      config->total_num_sources = num_strings;
    }
  }

  ret = TRUE;
done:
  if (error)
  {
    g_error_free(error);
  }
  if (keys)
  {
    g_strfreev(keys);
  }
  if (!ret)
  {
    NVGSTDS_ERR_MSG_V("%s failed", __func__);
  }
  return ret;
}

static gboolean
set_source_all_configs(NvDsConfig *config, gchar *cfg_file_path)
{
  guint i = 0;
  for (i = 0; i < config->total_num_sources; i++)
  {
    config->multi_source_config[i] = global_source_config;
    config->multi_source_config[i].camera_id = i;
    if (config->uri_list)
    {
      char *uri = config->uri_list[i];
      if (g_str_has_prefix(config->uri_list[i], "file://"))
      {
        config->multi_source_config[i].type = NV_DS_SOURCE_URI;
        config->multi_source_config[i].uri = g_strdup(uri + 7);
        config->multi_source_config[i].uri =
            g_strdup_printf("file://%s",
                            get_absolute_file_path(cfg_file_path,
                                                   config->multi_source_config[i].uri));
      }
      else if (g_str_has_prefix(config->uri_list[i], "rtsp://"))
      {
        config->multi_source_config[i].type = NV_DS_SOURCE_RTSP;
        config->multi_source_config[i].uri = config->uri_list[i];
      }
      else
      {
        gchar *source_id_start_ptr = uri + 4;
        gchar *source_id_end_ptr = NULL;
        long camera_id =
            g_ascii_strtoull(source_id_start_ptr, &source_id_end_ptr, 10);
        if (source_id_start_ptr == source_id_end_ptr || *source_id_end_ptr != '\0')
        {
          NVGSTDS_ERR_MSG_V("Incorrect URI for camera source %s. FORMAT: <usb/csi>:<dev_node/sensor_id>",
                            uri);
          return FALSE;
        }
        if (g_str_has_prefix(config->uri_list[i], "csi:"))
        {
          config->multi_source_config[i].type = NV_DS_SOURCE_CAMERA_CSI;
          config->multi_source_config[i].camera_csi_sensor_id = camera_id;
        }
        else if (g_str_has_prefix(config->uri_list[i], "usb:"))
        {
          config->multi_source_config[i].type = NV_DS_SOURCE_CAMERA_V4L2;
          config->multi_source_config[i].camera_v4l2_dev_node = camera_id;
        }
        else
        {
          NVGSTDS_ERR_MSG_V("URI %d (%s) not in proper format.", i,
                            config->uri_list[i]);
          return FALSE;
        }
      }
    }
  }
  return TRUE;
}

static gboolean
parse_app(NvDsConfig *config, GKeyFile *key_file, gchar *cfg_file_path)
{
  gboolean ret = FALSE;
  gchar **keys = NULL;
  gchar **key = NULL;
  GError *error = NULL;

  keys = g_key_file_get_keys(key_file, CONFIG_GROUP_APP, NULL, &error);
  CHECK_ERROR(error);

  for (key = keys; *key; key++)
  {
    if (!g_strcmp0(*key, CONFIG_GROUP_APP_ENABLE_PERF_MEASUREMENT))
    {
      config->enable_perf_measurement =
          g_key_file_get_integer(key_file, CONFIG_GROUP_APP,
                                 CONFIG_GROUP_APP_ENABLE_PERF_MEASUREMENT, &error);
      CHECK_ERROR(error);
    }
    else if (!g_strcmp0(*key, CONFIG_GROUP_APP_PERF_MEASUREMENT_INTERVAL))
    {
      config->perf_measurement_interval_sec =
          g_key_file_get_integer(key_file, CONFIG_GROUP_APP,
                                 CONFIG_GROUP_APP_PERF_MEASUREMENT_INTERVAL, &error);
      CHECK_ERROR(error);
    }
    else
    {
      NVGSTDS_WARN_MSG_V("Unknown key '%s' for group [%s]", *key,
                         CONFIG_GROUP_APP);
    }
  }

  ret = TRUE;
done:
  if (error)
  {
    g_error_free(error);
  }
  if (keys)
  {
    g_strfreev(keys);
  }
  if (!ret)
  {
    NVGSTDS_ERR_MSG_V("%s failed", __func__);
  }
  return ret;
}

gboolean
parse_config_file(NvDsConfig *config, gchar *cfg_file_path)
{
  GKeyFile *cfg_file = g_key_file_new();
  GError *error = NULL;
  gboolean ret = FALSE;
  gchar **groups = NULL;
  gchar **group;
  guint i, j;

  config->source_list_enabled = FALSE;

  if (!APP_CFG_PARSER_CAT)
  {
    GST_DEBUG_CATEGORY_INIT(APP_CFG_PARSER_CAT, "NVDS_CFG_PARSER", 0, NULL);
  }

  if (!g_key_file_load_from_file(cfg_file, cfg_file_path, G_KEY_FILE_NONE,
                                 &error))
  {
    GST_CAT_ERROR(APP_CFG_PARSER_CAT, "Failed to load uri file: %s",
                  error->message);
    goto done;
  }

  if (g_key_file_has_group(cfg_file, CONFIG_GROUP_SOURCE_LIST))
  {
    if (!parse_source_list(config, cfg_file, cfg_file_path))
    {
      GST_CAT_ERROR(APP_CFG_PARSER_CAT, "Failed to parse '%s' group",
                    CONFIG_GROUP_SOURCE_LIST);
      goto done;
    }
    config->num_source_sub_bins = config->total_num_sources;
    config->source_list_enabled = TRUE;
    if (!g_key_file_has_group(cfg_file, CONFIG_GROUP_SOURCE_ALL))
    {
      NVGSTDS_ERR_MSG_V("[source-attr-all] group not present.");
      ret = FALSE;
      goto done;
    }
    g_key_file_remove_group(cfg_file, CONFIG_GROUP_SOURCE_LIST, &error);
  }
  if (g_key_file_has_group(cfg_file, CONFIG_GROUP_SOURCE_ALL))
  {
    if (!parse_source(&global_source_config,
                      cfg_file, CONFIG_GROUP_SOURCE_ALL, cfg_file_path))
    {
      GST_CAT_ERROR(APP_CFG_PARSER_CAT, "Failed to parse '%s' group",
                    CONFIG_GROUP_SOURCE_LIST);
      goto done;
    }
    if (!set_source_all_configs(config, cfg_file_path))
    {
      ret = FALSE;
      goto done;
    }
    g_key_file_remove_group(cfg_file, CONFIG_GROUP_SOURCE_ALL, &error);
  }

  groups = g_key_file_get_groups(cfg_file, NULL);
  for (group = groups; *group; group++)
  {
    gboolean parse_err = FALSE;
    GST_CAT_DEBUG(APP_CFG_PARSER_CAT, "Parsing group: %s", *group);
    if (!g_strcmp0(*group, CONFIG_GROUP_APP))
    {
      parse_err = !parse_app(config, cfg_file, cfg_file_path);
    }

    if (!strncmp(*group, CONFIG_GROUP_SOURCE,
                 sizeof(CONFIG_GROUP_SOURCE) - 1))
    {
      if (config->num_source_sub_bins == MAX_SOURCE_BINS)
      {
        NVGSTDS_ERR_MSG_V("App supports max %d sources", MAX_SOURCE_BINS);
        ret = FALSE;
        goto done;
      }
      gchar *source_id_start_ptr = *group + strlen(CONFIG_GROUP_SOURCE);
      gchar *source_id_end_ptr = NULL;
      guint index =
          g_ascii_strtoull(source_id_start_ptr, &source_id_end_ptr, 10);
      if (source_id_start_ptr == source_id_end_ptr || *source_id_end_ptr != '\0')
      {
        NVGSTDS_ERR_MSG_V("Source group \"[%s]\" is not in the form \"[source<%%d>]\"",
                          *group);
        ret = FALSE;
        goto done;
      }
      guint source_id = 0;
      if (config->source_list_enabled)
      {
        if (index >= config->total_num_sources)
        {
          NVGSTDS_ERR_MSG_V("Invalid source group index %d, index cannot exceed %d", index,
                            config->total_num_sources);
          ret = FALSE;
          goto done;
        }
        source_id = index;
        NVGSTDS_INFO_MSG_V("Some parameters to be overwritten for group [%s]",
                           *group);
      }
      else
      {
        source_id = config->num_source_sub_bins;
      }
      parse_err = !parse_source(&config->multi_source_config[source_id],
                                cfg_file, *group, cfg_file_path);
      if (config->source_list_enabled && config->multi_source_config[source_id].type ==
                                             NV_DS_SOURCE_URI_MULTIPLE)
      {
        NVGSTDS_ERR_MSG_V("MultiURI support not available if [source-list] is provided");
        ret = FALSE;
        goto done;
      }
      if (config->multi_source_config[source_id].enable && !config->source_list_enabled)
      {
        config->num_source_sub_bins++;
      }
    }

    if (!g_strcmp0(*group, CONFIG_GROUP_STREAMMUX))
    {
      parse_err = !parse_streammux(&config->streammux_config, cfg_file, cfg_file_path);
    }

    if (!g_strcmp0(*group, CONFIG_GROUP_OSD))
      parse_err = !parse_osd(&config->osd_config, cfg_file);

    if (!strncmp(*group, CONFIG_GROUP_SINK, sizeof(CONFIG_GROUP_SINK) - 1))
    {
      if (config->num_sink_sub_bins == MAX_SINK_BINS)
      {
        NVGSTDS_ERR_MSG_V("App supports max %d sinks", MAX_SINK_BINS);
        ret = FALSE;
        goto done;
      }
      parse_err =
          !parse_sink(&config->sink_bin_sub_bin_config[config->num_sink_sub_bins], cfg_file, *group,
                      cfg_file_path);
      if (config->sink_bin_sub_bin_config[config->num_sink_sub_bins].enable)
      {
        config->num_sink_sub_bins++;
      }
    }

    if (parse_err)
    {
      GST_CAT_ERROR(APP_CFG_PARSER_CAT, "Failed to parse '%s' group", *group);
      goto done;
    }
  }

  /* Updating batch size when source list is enabled */
  if (config->source_list_enabled == TRUE)
  {
    /* For streammux and pgie, batch size is set to number of sources */
    config->streammux_config.batch_size = config->num_source_sub_bins;
  }

  for (i = 0; i < config->num_source_sub_bins; i++)
  {
    if (config->multi_source_config[i].type == NV_DS_SOURCE_URI_MULTIPLE)
    {
      if (config->multi_source_config[i].num_sources < 1)
      {
        config->multi_source_config[i].num_sources = 1;
      }
      for (j = 1; j < config->multi_source_config[i].num_sources; j++)
      {
        if (config->num_source_sub_bins == MAX_SOURCE_BINS)
        {
          NVGSTDS_ERR_MSG_V("App supports max %d sources", MAX_SOURCE_BINS);
          ret = FALSE;
          goto done;
        }
        memcpy(&config->multi_source_config[config->num_source_sub_bins],
               &config->multi_source_config[i],
               sizeof(config->multi_source_config[i]));
        config->multi_source_config[config->num_source_sub_bins].type =
            NV_DS_SOURCE_URI;
        config->multi_source_config[config->num_source_sub_bins].uri =
            g_strdup_printf(config->multi_source_config[config->num_source_sub_bins].uri, j);
        config->num_source_sub_bins++;
      }
      config->multi_source_config[i].type = NV_DS_SOURCE_URI;
      config->multi_source_config[i].uri =
          g_strdup_printf(config->multi_source_config[i].uri, 0);
    }
  }
  ret = TRUE;

done:
  if (cfg_file)
    g_key_file_free(cfg_file);

  if (groups)
    g_strfreev(groups);

  if (error)
    g_error_free(error);

  if (!ret)
    NVGSTDS_ERR_MSG_V("%s failed", __func__);

  return ret;
}
