#include "main.h"

AppCtx *appCtx;                     
static gchar *cfg_file_path = NULL; 
static GOptionContext *ctx;         
static GMainLoop *main_loop = NULL; 
                                    
                                    
static gint return_value = 0;       
static guint cintr = FALSE;
static gboolean quit = FALSE;

GST_DEBUG_CATEGORY(NVDS_APP);

/**
 * Function to handle program interrupt signal.
 * It installs default handler after handling the interrupt.
 */
static void
_intr_handler(int signum)
{
    struct sigaction action;
    NVGSTDS_ERR_MSG_V("User Interrupted.. \n");
    memset(&action, 0, sizeof(action));
    action.sa_handler = SIG_DFL;
    sigaction(SIGINT, &action, NULL);
    cintr = TRUE;
}

/*
 * Function to install custom handler for program interrupt signal.
 */
static void
_intr_setup(void)
{
    struct sigaction action;
    memset(&action, 0, sizeof(action));
    action.sa_handler = _intr_handler;
    sigaction(SIGINT, &action, NULL);
}

/**
 * Loop function to check the status of interrupts.
 * It comes out of loop if application got interrupted.
 */
static gboolean
check_for_interrupt(gpointer data)
{
    if (quit)
        return FALSE;

    if (cintr)
    {
        cintr = FALSE;
        quit = TRUE;
        g_main_loop_quit(main_loop);
        return FALSE;
    }
    return TRUE;
}

static void parse_args(int argc, char *argv[])
{
    GError *error = NULL;

    GOptionEntry entries[] = {
        {"cfg-file", 'c', 0, G_OPTION_ARG_FILENAME, &cfg_file_path,
         "Set the config file", NULL},
        {NULL},
    };

    ctx = g_option_context_new("Nvidia DeepStream Demo");
    GOptionGroup *group = g_option_group_new("abc", NULL, NULL, NULL, NULL);
    g_option_group_add_entries(group, entries);

    g_option_context_set_main_group(ctx, group);
    g_option_context_add_group(ctx, gst_init_get_option_group());

    GST_DEBUG_CATEGORY_INIT(NVDS_APP, "NVDS_APP", 0, NULL);

    if (!g_option_context_parse(ctx, &argc, &argv, &error))
    {
        NVGSTDS_ERR_MSG_V("%s", error->message);
        return_value = -1;
    }

    if (!cfg_file_path)
    {
        NVGSTDS_ERR_MSG_V("Specify config file with -c option");
        return_value = -1;
    }
}

int main(int argc, char *argv[])
{
    gst_init(&argc, &argv);
    NVGSTDS_INFO_MSG_V("Initialization of sova deepstream.\n");

    parse_args(argc, argv);
    if (return_value == -1)
        goto done;

    appCtx = g_malloc0(sizeof(AppCtx));

    if (!parse_config_file(&appCtx->config, cfg_file_path))
    {
        NVGSTDS_ERR_MSG_V("Failed to parse config file '%s'", cfg_file_path);
        return_value = -1;
        goto done;
    }

    if (!create_pipeline(appCtx))
    {
        NVGSTDS_ERR_MSG_V("Failed to create pipeline");
        return_value = -1;
        goto done;
    }

    main_loop = g_main_loop_new(NULL, FALSE);

    _intr_setup();
    g_timeout_add(400, check_for_interrupt, NULL);

    if (gst_element_set_state(appCtx->pipeline.pipeline,
                              GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE)
    {
        NVGSTDS_ERR_MSG_V("Can't set pipeline to playing state.\n");
        return_value = -1;
        goto done;
    }

    g_main_loop_run(main_loop);

done:
    NVGSTDS_INFO_MSG_V("Quitting\n");

    destroy_pipeline(appCtx);
    g_free(appCtx);

    if (main_loop)
        g_main_loop_unref(main_loop);
    if (ctx)
        g_option_context_free(ctx);

    if (return_value == 0)
        NVGSTDS_INFO_MSG_V("App run successful\n");
    else
        NVGSTDS_INFO_MSG_V("App run failed\n");

    gst_deinit();

    return return_value;
}