#include "bins.h"
#include "../metadata.h"
#include "../utils.h"

gboolean
create_sink_instance_bin(NvDsInstanceBin *instance_bin, NvDsOSDConfig *osd_config,
                         NvDsSinkSubBinConfig *sink_bin_sub_bin_config_array,
                         guint num_sink_sub_bins)
{
    gboolean ret = FALSE;

    gchar elem_name[32];
    g_snprintf(elem_name, 32, "sink_bin_0");
    instance_bin->bin = gst_bin_new(elem_name);

    // Create sink bin to write stream in a file. This is standart function from deepstream_sinks.h
    if (!create_sink_bin(num_sink_sub_bins, sink_bin_sub_bin_config_array, &instance_bin->sink_bin, 0))
        goto done;
    gst_bin_add(GST_BIN(instance_bin->bin), instance_bin->sink_bin.bin);

    // osd - on screen display - show information on video stream. For example draw bounding
    // boxes of detected people. OSD information is processing in process_metadata probe
    // function in metadata.c file.
    if (osd_config->enable)
    {
        if (!create_osd_bin(osd_config, &instance_bin->osd_bin))
            goto done;
        gst_bin_add(GST_BIN(instance_bin->bin), instance_bin->osd_bin.bin);

        NVGSTDS_LINK_ELEMENT(instance_bin->osd_bin.bin, instance_bin->sink_bin.bin);
    }

    if (osd_config->enable)
        // Add ghost sink pad on instance_bin->bin based on osd_bin.bin sink pad.
        NVGSTDS_BIN_ADD_GHOST_PAD(instance_bin->bin, instance_bin->osd_bin.bin, "sink");
    else
        // Add ghost sink pad on instance_bin->bin based on sink_bin.bin sink pad.
        NVGSTDS_BIN_ADD_GHOST_PAD(instance_bin->bin, instance_bin->sink_bin.bin, "sink");

    ret = TRUE;
done:
    if (!ret)
        NVGSTDS_ERR_MSG_V("%s failed", __func__);

    return ret;
}

/**
 * @param pipeline - main pipeline structure
 * @param index - index of sink bin
 * @param appCtx - main structure of app that contains all the information about the application.
 * @return gboolean returns TRUE - if everything is OK, else FALSE
 */
static gboolean create_instance_tee(NvDsPipeline *pipeline, AppCtx *appCtx)
{
    gboolean ret = FALSE;
    NvDsInstanceBin *instance_bin = &pipeline->instance_bins[0];

    // If tee has not been created yet
    if (!instance_bin->tee)
    {
        // Init tee and add to pipeline
        gchar tee_name[30];
        g_snprintf(tee_name, 30, "instance_bin_tee");
        instance_bin->tee = gst_element_factory_make(NVDS_ELEM_TEE, tee_name);
        if (!instance_bin->tee)
        {
            NVGSTDS_ERR_MSG_V("Failed to create element '%s'", tee_name);
            goto done;
        }
        gst_bin_add(GST_BIN(pipeline->pipeline), instance_bin->tee);

        gchar pad_name[16];
        g_snprintf(pad_name, 16, "src_%02d", 0);
        GstPad *demux_src_pad = gst_element_get_request_pad(pipeline->demuxer, pad_name);
        NVGSTDS_LINK_ELEMENT_FULL(pipeline->demuxer, pad_name, instance_bin->tee, "sink");
        gst_object_unref(demux_src_pad);

        // Add probe function to draw information on stream using OSD.
        NVGSTDS_ELEM_ADD_PROBE(instance_bin->all_bbox_buffer_probe_id,
                               instance_bin->tee, "sink",
                               process_metadata, GST_PAD_PROBE_TYPE_BUFFER, appCtx);
    }

    ret = TRUE;
done:
    if (!ret)
        NVGSTDS_ERR_MSG_V("%s failed", __func__);

    return ret;
}

gboolean create_sink(NvDsPipeline *pipeline, NvDsConfig *config, AppCtx *appCtx)
{
    gboolean ret = FALSE;

    // Create filename for sink file.
    gchar time_string[30];
    time_t now = time(NULL);
    strftime(time_string, 22, "%Y-%m-%d__%H_%M_%S", localtime(&now));
    gchar *file_name = (char *)malloc(255 * sizeof(char));
    sprintf(file_name, "video_logs/data/%s_sink.mp4", time_string);
    config->sink_bin_sub_bin_config[0].encoder_config.output_file_path = file_name;

    // Create sink instance bin to write stream in a file.
    if (!create_sink_instance_bin(&pipeline->instance_bins[0], &config->osd_config,
                                  config->sink_bin_sub_bin_config, config->num_sink_sub_bins))
        goto done;
    g_object_set(G_OBJECT(pipeline->instance_bins[0].bin), "message-forward", TRUE, NULL);
    gst_bin_add(GST_BIN(pipeline->pipeline), pipeline->instance_bins[0].bin);

    if (!create_instance_tee(pipeline, appCtx))
        goto done;

    // Link tee pad with instance_bin pad
    NVGSTDS_LINK_ELEMENT(pipeline->instance_bins[0].tee, pipeline->instance_bins[0].bin);

    ret = TRUE;
done:
    if (!ret)
        NVGSTDS_ERR_MSG_V("%s failed", __func__);

    return ret;
}
