#include "metadata.h"
#include "main.h"
#include "utils.h"
#include "../video_logs/video_logs.h"

GST_DEBUG_CATEGORY_EXTERN(NVDS_APP);

gboolean bus_callback(GstBus *bus, GstMessage *message, gpointer data)
{
  AppCtx *appCtx = (AppCtx *)data;
  GST_CAT_DEBUG(NVDS_APP,
                "Received message on bus: source %s, msg_type %s",
                GST_MESSAGE_SRC_NAME(message), GST_MESSAGE_TYPE_NAME(message));
  switch (GST_MESSAGE_TYPE(message))
  {
  case GST_MESSAGE_INFO:
  {
    GError *error = NULL;
    gchar *debuginfo = NULL;
    gst_message_parse_info(message, &error, &debuginfo);
    NVGSTDS_INFO_MSG_V("INFO from %s: %s\n",
              GST_OBJECT_NAME(message->src), error->message);
    if (debuginfo)
    {
      NVGSTDS_INFO_MSG_V("Debug info: %s\n", debuginfo);
    }
    g_error_free(error);
    g_free(debuginfo);
    break;
  }
  case GST_MESSAGE_WARNING:
  {
    GError *error = NULL;
    gchar *debuginfo = NULL;
    gst_message_parse_warning(message, &error, &debuginfo);
    NVGSTDS_WARN_MSG_V("WARNING from %s: %s\n",
              GST_OBJECT_NAME(message->src), error->message);
    if (debuginfo)
    {
      NVGSTDS_INFO_MSG_V("Debug info: %s\n", debuginfo);
    }
    g_error_free(error);
    g_free(debuginfo);
    break;
  }
  case GST_MESSAGE_ERROR:
  {
    GError *error = NULL;
    gchar *debuginfo = NULL;
    guint i = 0;
    gst_message_parse_error(message, &error, &debuginfo);
    NVGSTDS_ERR_MSG_V("ERROR from %s: %s\n",
             GST_OBJECT_NAME(message->src), error->message);
    if (debuginfo)
    {
      NVGSTDS_INFO_MSG_V("Debug info: %s\n", debuginfo);
    }

    NvDsSrcParentBin *bin = &appCtx->pipeline.multi_src_bin;
    GstElement *msg_src_elem = (GstElement *)GST_MESSAGE_SRC(message);
    gboolean bin_found = FALSE;
    /* Find the source bin which generated the error. */
    while (msg_src_elem && !bin_found)
    {
      for (i = 0; i < bin->num_bins && !bin_found; i++)
      {
        if (bin->sub_bins[i].src_elem == msg_src_elem ||
            bin->sub_bins[i].bin == msg_src_elem)
        {
          bin_found = TRUE;
          break;
        }
      }
      msg_src_elem = GST_ELEMENT_PARENT(msg_src_elem);
    }

    if ((i != bin->num_bins) &&
        (appCtx->config.multi_source_config[0].type == NV_DS_SOURCE_RTSP))
    {
      // Error from one of RTSP source.
      NvDsSrcBin *subBin = &bin->sub_bins[i];

      if (!subBin->reconfiguring ||
          g_strrstr(debuginfo, "500 (Internal Server Error)"))
      {
        subBin->reconfiguring = TRUE;
        g_timeout_add(0, reset_source_pipeline, subBin);
      }
      g_error_free(error);
      g_free(debuginfo);
      return TRUE;
    }

    if (appCtx->config.multi_source_config[0].type == NV_DS_SOURCE_CAMERA_V4L2)
    {
      if (g_strrstr(debuginfo, "reason not-negotiated (-4)"))
      {
        NVGSTDS_INFO_MSG_V("incorrect camera parameters provided, please provide supported resolution and frame rate\n");
      }

      if (g_strrstr(debuginfo, "Buffer pool activation failed"))
      {
        NVGSTDS_INFO_MSG_V("usb bandwidth might be saturated\n");
      }
    }

    g_error_free(error);
    g_free(debuginfo);
    break;
  }
  case GST_MESSAGE_STATE_CHANGED:
  {
    GstState oldstate, newstate;
    gst_message_parse_state_changed(message, &oldstate, &newstate, NULL);
    if (GST_ELEMENT(GST_MESSAGE_SRC(message)) == appCtx->pipeline.pipeline)
    {
      switch (newstate)
      {
      case GST_STATE_PLAYING:
        NVGSTDS_INFO_MSG_V("Pipeline is running\n");
        GST_DEBUG_BIN_TO_DOT_FILE_WITH_TS(GST_BIN(appCtx->pipeline.pipeline), GST_DEBUG_GRAPH_SHOW_ALL,
                                          "ds-app-playing");
        break;
      case GST_STATE_PAUSED:
        if (oldstate == GST_STATE_PLAYING)
        {
          NVGSTDS_INFO_MSG_V("Pipeline paused\n");
        }
        break;
      case GST_STATE_READY:
        GST_DEBUG_BIN_TO_DOT_FILE_WITH_TS(GST_BIN(appCtx->pipeline.pipeline), GST_DEBUG_GRAPH_SHOW_ALL, "ds-app-ready");
        if (oldstate == GST_STATE_NULL)
        {
          NVGSTDS_INFO_MSG_V("Pipeline ready\n");
        }
        else
        {
          NVGSTDS_INFO_MSG_V("Pipeline stopped\n");
        }
        break;
      case GST_STATE_NULL:
        g_mutex_lock(&appCtx->app_lock);
        g_cond_broadcast(&appCtx->app_cond);
        g_mutex_unlock(&appCtx->app_lock);
        break;
      default:
        break;
      }
    }
    break;
  }
  case GST_MESSAGE_ELEMENT:
  {
    const GstStructure *s = gst_message_get_structure(message);
    if (gst_structure_has_name(s, "GstBinForwarded"))
    {
      GstMessage *forward_msg = NULL;

      gst_structure_get(s, "message", GST_TYPE_MESSAGE, &forward_msg, NULL);
      if (GST_MESSAGE_TYPE(forward_msg) == GST_MESSAGE_EOS)
      {
        NVGSTDS_INFO_MSG_V("EOS from element %s\n", GST_OBJECT_NAME(GST_MESSAGE_SRC(forward_msg)));

        NvDsPipeline *pipeline = &appCtx->pipeline;
        NvDsInstanceBin *instance_bin = &pipeline->instance_bins[0];
        GstPad *blockpad;

        // Remove sink bin
        gst_element_set_state(instance_bin->bin, GST_STATE_NULL);
        gst_bin_remove(GST_BIN(pipeline->pipeline), instance_bin->bin);

        // Create and to pipeline fakesink
        instance_bin->fake_sink = gst_element_factory_make("fakesink", "fakesink");
        gst_bin_add(GST_BIN(pipeline->pipeline), instance_bin->fake_sink);
        gst_element_link(instance_bin->tee, instance_bin->fake_sink);
        gst_element_set_state(instance_bin->fake_sink, GST_STATE_PLAYING);

        // Remove blocking probe pad.
        blockpad = gst_element_get_static_pad(instance_bin->tee, "sink");
        gst_pad_remove_probe(blockpad, instance_bin->probe_id);
        gst_object_unref(blockpad);
      }
      gst_message_unref(forward_msg);
    }
    break;
  }
  default:
    break;
  }
  return TRUE;
}

GstPadProbeReturn process_metadata(GstPad *pad, GstPadProbeInfo *info, gpointer u_data)
{
  GstBuffer *buf = (GstBuffer *)info->data;
  AppCtx *appCtx = (AppCtx *)u_data;

  if (gst_buffer_is_writable(buf))
  {
    NvDsBatchMeta *batch_meta = gst_buffer_get_nvds_batch_meta(buf);
    if (!batch_meta)
    {
      NVGSTDS_WARN_MSG_V("Batch meta not found for buffer %p", buf);
      return GST_PAD_PROBE_OK;
    }
    video_logger(appCtx);
  }
  return GST_PAD_PROBE_OK;
}
