#include "bins/bins.h"
#include "metadata.h"
#include "utils.h"

gboolean create_pipeline(AppCtx *appCtx)
{
    gboolean ret = FALSE;
    NvDsPipeline *pipeline = &appCtx->pipeline;
    NvDsConfig *config = &appCtx->config;
    GstElement *lastElem;

    pipeline->pipeline = gst_pipeline_new("pipeline");

    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    pipeline->bus_id = gst_bus_add_watch(bus, bus_callback, appCtx);
    gst_object_unref(bus);

    if (!create_source(pipeline, config, &lastElem))
        goto done;

    if (!create_demuxer(pipeline, config, lastElem))
        goto done;

    if (!create_sink(pipeline, config, appCtx))
        goto done;

    if (config->enable_perf_measurement)
    {
        GstPad *fps_pad;
        fps_pad = gst_element_get_static_pad(pipeline->demuxer, "sink");

        appCtx->perf_struct.context = appCtx;
        enable_perf_measurement(&appCtx->perf_struct, fps_pad,
                                pipeline->multi_src_bin.num_bins,
                                config->perf_measurement_interval_sec,
                                config->multi_source_config[0].dewarper_config.num_surfaces_per_frame,
                                perf_cb);
        gst_object_unref(fps_pad);
    }

    g_mutex_init(&appCtx->app_lock);
    g_cond_init(&appCtx->app_cond);

    ret = TRUE;
done:
    if (!ret)
        NVGSTDS_ERR_MSG_V("%s failed", __func__);

    return ret;
}

void destroy_pipeline(AppCtx *appCtx)
{
    NvDsConfig *config = &appCtx->config;
    GstBus *bus = NULL;
    gint64 end_time = g_get_monotonic_time() + G_TIME_SPAN_SECOND;

    if (!appCtx)
        return;

    if (appCtx->pipeline.demuxer)
    {
        gst_pad_send_event(
            gst_element_get_static_pad(appCtx->pipeline.demuxer, "sink"), gst_event_new_eos());
    }
    else if (appCtx->pipeline.instance_bins[0].sink_bin.bin)
    {
        gst_pad_send_event(
            gst_element_get_static_pad(appCtx->pipeline.instance_bins[0].sink_bin.bin, "sink"),
            gst_event_new_eos());
    }

    g_usleep(100000);

    g_mutex_lock(&appCtx->app_lock);
    if (appCtx->pipeline.pipeline)
    {
        destroy_smart_record_bin(&appCtx->pipeline.multi_src_bin);
        bus = gst_pipeline_get_bus(GST_PIPELINE(appCtx->pipeline.pipeline));

        while (TRUE)
        {
            GstMessage *message = gst_bus_pop(bus);
            if (message == NULL)
                break;
            else if (GST_MESSAGE_TYPE(message) == GST_MESSAGE_ERROR)
                bus_callback(bus, message, appCtx);
            else
                gst_message_unref(message);
        }
        gst_element_set_state(appCtx->pipeline.pipeline, GST_STATE_NULL);
    }
    g_cond_wait_until(&appCtx->app_cond, &appCtx->app_lock, end_time);
    g_mutex_unlock(&appCtx->app_lock);

    for (guint i = 0; i < appCtx->config.num_source_sub_bins; i++)
    {
        NvDsInstanceBin *bin = &appCtx->pipeline.instance_bins[i];
        NVGSTDS_ELEM_REMOVE_PROBE(bin->all_bbox_buffer_probe_id, bin->tee, "sink");
    }
    destroy_sink_bin();

    if (appCtx->pipeline.pipeline)
    {
        bus = gst_pipeline_get_bus(GST_PIPELINE(appCtx->pipeline.pipeline));
        gst_bus_remove_watch(bus);
        gst_object_unref(bus);
        gst_object_unref(appCtx->pipeline.pipeline);
    }
}