#include "video_logs.h"
#include "../pipeline/bins/bins.h"
#include "../pipeline/utils.h"
#include "../pipeline/metadata.h"

#define MAX_FILE_DURATION 3

time_t stream_time_created = 0;
time_t cur_time;

/**
 * @brief Probe pad callback function, which remove sink_bin from pipeline
 * and put fakesink instead of it.
 *
 * @param pad - a pad on which the probe is set.
 * @param info
 * @param user_data - data, we sent here ProbeData instance.
 * @return GstPadProbeReturn
 */
static GstPadProbeReturn
remove_sink_probe_cb(GstPad *pad, GstPadProbeInfo *info, gpointer user_data)
{
    // Get necessary variables
    AppCtx *appCtx = (AppCtx *)user_data;

    // Send end of stream signal to sink bin and flush all data to the file.
    GstPad *sink_pad = gst_element_get_static_pad(
        appCtx->pipeline.instance_bins[0].bin, "sink");
    gst_pad_send_event(sink_pad, gst_event_new_eos());
    gst_object_unref(sink_pad);

    return GST_PAD_PROBE_OK;
}

/**
 * @brief Create block probe with remove_sink_probe_cb function callback
 */
static void create_probe_remove(NvDsInstanceBin *instance_bin, AppCtx *appCtx)
{
    GstPad *blockpad = gst_element_get_static_pad(instance_bin->tee, "sink");
    if (!gst_pad_is_blocked(blockpad))
        instance_bin->probe_id = gst_pad_add_probe(blockpad, GST_PAD_PROBE_TYPE_BLOCK_DOWNSTREAM,
                                                   remove_sink_probe_cb, appCtx, NULL);
    gst_object_unref(blockpad);
}

/**
 * @brief Probe pad callback function, which create new sink_bin and remove
 * fakesink from pipeline.
 *
 * @param pad - a pad on which the probe is set.
 * @param info
 * @param user_data - data, we sent appCtx instance here.
 * @return GstPadProbeReturn
 */
static GstPadProbeReturn
create_sink_probe_cb(GstPad *pad, GstPadProbeInfo *info, gpointer user_data)
{
    AppCtx *appCtx = (AppCtx *)user_data;
    NvDsPipeline *pipeline = &appCtx->pipeline;
    NvDsConfig *config = &appCtx->config;
    NvDsInstanceBin *instance_bin = &pipeline->instance_bins[0];

    // Remove fakesink
    if (instance_bin->fake_sink)
    {
        gst_element_set_state(instance_bin->fake_sink, GST_STATE_NULL);
        gst_bin_remove(GST_BIN(pipeline->pipeline), instance_bin->fake_sink);
        instance_bin->fake_sink = NULL;
    }

    // Create new sink bin
    if (!create_sink(pipeline, config, appCtx))
        NVGSTDS_ERR_MSG_V("Error to change fakesink to real sink bin\n");
    if (!gst_element_sync_state_with_parent(instance_bin->bin))
        NVGSTDS_ERR_MSG_V("Unable to set state as parent\n");

    return GST_PAD_PROBE_REMOVE;
}

void video_logger(AppCtx *appCtx)
{
    NvDsPipeline *pipeline = &appCtx->pipeline;
    NvDsConfig *config = &appCtx->config;
    time(&cur_time);

    // Initialization
    if (stream_time_created == 0)
        time(&stream_time_created);

    // If fakesink is used than change it to real sink
    if (pipeline->instance_bins[0].fake_sink)
    {
        stream_time_created = cur_time;

        GstPad *blockpad = gst_element_get_static_pad(pipeline->instance_bins[0].tee, "sink");
        if (!gst_pad_is_blocked(blockpad))
            gst_pad_add_probe(blockpad, GST_PAD_PROBE_TYPE_BLOCK_DOWNSTREAM,
                              create_sink_probe_cb, appCtx, NULL);
        gst_object_unref(blockpad);
    }

    // Close file if it's more than max duration
    if (cur_time - stream_time_created >= MAX_FILE_DURATION)
        create_probe_remove(&pipeline->instance_bins[0], appCtx);
}